import os
import pathlib
import ctypes
from ctypes import c_double, c_int, c_bool, CFUNCTYPE, POINTER
import numpy as np
from numpy.ctypeslib import ndpointer

if os.name == 'nt':
    so_ext = ".pyd"
else:
    import distutils.ccompiler
    so_ext = distutils.ccompiler.new_compiler().shared_lib_extension

class _sdof_peaks(ctypes.Structure):
    _fields_ = [
        ("max_displ",      c_double),
        ("max_veloc",      c_double),
        ("max_accel",      c_double)
    ]

class _sdof_config(ctypes.Structure):
    _fields_ = [
        ("alpha_m",      c_double),
        ("alpha_f",      c_double),
        ("beta",         c_double),
        ("gamma",        c_double)
    ]


try:
    libfile = str(next(pathlib.Path(__file__).parents[0].glob("_fsdof.*"+so_ext)))
    lib = ctypes.cdll.LoadLibrary(libfile)
    # conf = lib.CONF
    _fsdof_peaks = lib.fsdof_peaks
    _fsdof_peaks.restype = c_int
    _fsdof_peaks.argtypes = (
        POINTER(_sdof_config),
        c_double,  c_double,  c_double,
        c_double, c_int, POINTER(c_double), c_double,
        POINTER(_sdof_peaks)
    )

    _fsdof_integrate = lib.fsdof_integrate
    _fsdof_integrate.restype = c_int
    _fsdof_integrate.argtypes = (
        POINTER(_sdof_config),
        c_double,  c_double,  c_double,
        c_double, c_int, POINTER(c_double), c_double,
        ndpointer(c_double, flags="C_CONTIGUOUS")
    )

    _fsdof_integrate2 = lib.fsdof_integrate2
    _fsdof_integrate2.restype = c_int
    _fsdof_integrate2.argtypes = (
        POINTER(_sdof_config),
        c_double,  c_double,  c_double,
        c_double, c_int, POINTER(c_double), c_double,
        ndpointer(c_double, flags="C_CONTIGUOUS")
    )

    CONFIG = _sdof_config()

except:
    raise

#   elastic_sdof()
#   plastic


def integrate(m,c,k,f,dt, u0=0.0, v0=0.0,
              out  =  None,
              alpha_m: float = 1.0,
              alpha_f: float = 1.0,
              beta   : float = 0.25,
              gamma  : float = 0.50
    ):

    import numpy as np
    if out is None:
        output = np.empty((3,len(f)))
    else:
        output = out
    output[:2,0] = u0, v0

    config = _sdof_config(
                alpha_m = alpha_m,
                alpha_f = alpha_f,
                beta    = beta,
                gamma   = gamma
    )

    _fsdof_integrate(config, m, c, k, 1.0, len(f), np.asarray(f).ctypes.data_as(POINTER(c_double)), dt, output)
    return output

def integrate2(m,c,k,f,dt, u0=0.0, v0=0.0,
              out  =  None,
              alpha_m: float = 1.0,
              alpha_f: float = 1.0,
              beta   : float = 0.25,
              gamma  : float = 0.50
    ):
    import numpy as np
    if out is None:
        output = np.empty((3,len(f)))
    else:
        output = out
    output[0,:2] = u0, v0

    config = _sdof_config(
                alpha_m=alpha_m,
                alpha_f=alpha_f,
                beta   =beta,
                gamma  =gamma
    )

    _fsdof_integrate2(CONFIG, m, c, k, 1.0, len(f), np.asarray(f).ctypes.data_as(POINTER(c_double)), dt, output)
    return output.T



def _thread_spectrum(n_threads=1):
    import threading
    threads = []
    for i in range(n_threads):
        threads.append(
                threading.Thread(
                    target=some_ctypes_func, args=(arg, )
                )
        )

    for thread in threads:
        thread.start()

    for thread in threads:
        thread.join()


def spectrum(accel, dt, damping, periods=None, interp=None, **kwds):

    if interp is None:
        from scipy.interpolate import interp1d as interp

    if isinstance(damping, float):
        damping = [damping]

    if periods is None:
        periods = np.linspace(0.02, 3.0, 200)
    elif isinstance(periods, tuple):
        periods = np.linspace(*periods)
    elif isinstance(periods, list):
        periods = np.array(periods)


    pi = np.pi
    mass = 1.0
    numdata = len(accel)
    t       = np.arange(0,(numdata)*dt,dt)
    t_max   = max(t)

    u0, v0 = 0.0, 0.0

    Sd, Sv, Sa = np.zeros((3,1+len(damping), len(periods)))
    Sd[0,:] = periods[:]
    Sv[0,:] = periods[:]
    Sa[0,:] = periods[:]

    for i,dmp in enumerate(damping):
        for j,period in enumerate(periods):
            if dt/period > 0.02:
                dtp = period*0.02
                dtpx = np.arange(0,t_max,dtp)
                dtpx = dtpx
                accfrni = interp(t, accel)(dtpx)
                accfrn = accfrni[1:len(accfrni)-1]
                numdatan = len(accfrn)
            else:
                dtp = dt
                accfrn = accel
                numdatan = numdata

            p = -mass*accfrn
            k = 4*pi**2*mass/period**2
            c = 2*dmp*np.sqrt(k/mass)

            d, v, a = integrate(mass, c, k, p, dt)

            # resp = peaks(m, c, k, p, dt)
            # resp.max_displ
            # resp.max_accel

            Sd[1+i,j] = abs(max(d, key=abs))
            Sv[1+i,j] = abs(max(v, key=abs))
            Sa[1+i,j] = abs(max(a+accfrn, key=abs))

    return Sd,Sv,Sa



def peaks(m,c,k, f, dt):
    response = _sdof_peaks()
    _fsdof_peaks(CONFIG, m, c, k, 1.0, len(f), f.ctypes.data_as(POINTER(c_double)), dt, response)
    return response




import sdof
import numpy as np


print(sdof.peaks(
        1., 1., 0.,
        np.loadtxt("data/elCentro.txt"), 0.01
).max_accel)

f =     np.array([
         0.0000,
         5.0000,
         8.6603,
        10.0000,
         8.6603,
         5.0000,
         0.0000,
         0.0000,
         0.0000,
         0.0000,
         0.0000])

from pandas import DataFrame
U = sdof.integrate(0.2533, 0.1592, 10., f, 0.10)
print(DataFrame(U.T))

import matplotlib.pyplot as plt

u,v,a = sdof.integrate(0.2533, 0.0, 10., np.zeros(100), 0.01, u0=1.0)

fig, ax = plt.subplots()
# ax.plot(a)
# ax.plot(v)
ax.plot(u)
plt.show()

u,v,a = sdof.integrate(
        1., 1.e3, 0.,
        -np.sin(np.linspace(0, 5*np.pi, 200)), 5*np.pi/200
)

fig, ax = plt.subplots()
ax.plot(a)
ax.plot(v)
ax.plot(u)
plt.show()

f = np.loadtxt("data/elCentro.txt")
u,v,a = sdof.integrate(
        1., 1., 0.,
        f, 0.01
)
fig, ax = plt.subplots()
ax.plot(a)
ax.plot(v)
ax.plot(u)
ax.plot(f)
plt.show()




import numpy as np
from sdof import spectrum
"""
# Response Spectrum
elceltro | sdof $dt
elcentro | sdof $dt $damp
elcentro | sdof $dt $damp $period
sdof input.txt $dt
sdof input.txt $dt $damp
sdof input.txt $dt $damp $period

# Response History
"""

def parse_args(args):
    options = {
        "include_columns": ("periods",  "accel"), #"pseudo_accel",
        "return_mode": "spectra"
    }
    return options

COLUMNS = {
    "periods":      lambda Sd, *_:  Sd[0,:][:,None],
    "pseudo_accel": lambda Sd, *_: (Sd[1:,:]*(2*np.pi/Sd[0,:])**2).T,
    "accel":        lambda Sd, Sv, Sa: Sa[1:,:].T
}


if __name__ == "__main__":
    import sys
    options = parse_args(sys.argv)

    accel  = np.loadtxt(sys.argv[1]) if len(sys.argv) > 1 else np.loadtxt(sys.stdin)
    dt     = float(sys.argv[-1])

    # Response Spectra mode

    Sd, Sv, Sa = spectrum(accel, dt, [0.02, 0.05], None)
    stride = 2
    output = np.zeros((len(Sd[0,:]),stride*(len(options["include_columns"])-1)+1))

    output[:,0] = Sd[0,:]

    for i, key in enumerate(options["include_columns"][1:]):
        resp   = COLUMNS[key](Sd, Sv, Sa)
        # stride = resp.shape[1] if len(resp.shape) == 2 else 1

        output[:,i*stride+1:(i+1)*stride+1] = resp #(Sd[1:,:]*(2*np.pi/Sd[0,:])**2).T


    np.savetxt(sys.stdout.buffer, output)





from setuptools import Extension, setup

setup(
    ext_modules=[
        Extension(
            name="sdof._fsdof",
            sources=["src/fsdof.c"],
            export_symbols=[]
        ),
    ]
)
